#include "sdk_config.h"

#include "nrf_vt_backend.h"
#include "nrf_vt_constants.h"

#include "nrf_error.h"
#include "nrfx_log.h"
#include "nrfx_saadc.h"
#include "nrfx_pwm.h"
#include "nrf_gpio.h"
#include "nrf_ppi.h"

#include "SEGGER_RTT.h"
#include "app_scheduler.h"

#if defined(ADC_SAMPLER_PROFILING) || defined(PWM_TRANSMITTER_PROFILING)
#include "nrf_gpiote.h"
#endif

#if defined(NRF_VT_ENABLED) && NRF_VT_ENABLED


/*****************************************************************************
 * Constants
 *****************************************************************************/

/*
 * ADC constants / macros
 *****************************************************************************/

/** The resolution used for the ADC */
#define ADC_SAMPLER_RESOLUTION  (NRF_SAADC_RESOLUTION_10BIT)
/** The oversample setting used for the ADC */
#define ADC_SAMPLER_OVERSAMPLE  (NRF_SAADC_OVERSAMPLE_DISABLED)
/** The Low Power mode used for the nrfx_saadc library */
#define ADC_SAMPLER_LP_MODE     (false)

/** The resistor setting used for the positive ADC input */
#define ADC_SAMPLER_RESISTOR_P  (NRF_SAADC_RESISTOR_PULLDOWN)
/** The resistor setting used for the negative ADC input */
#define ADC_SAMPLER_RESISTOR_N  (NRF_SAADC_RESISTOR_DISABLED)
/** The gain setting used for the ADC */
#define ADC_SAMPLER_GAIN        (NRF_SAADC_GAIN1_6)
/** The reference setting used for the ADC */
#define ADC_SAMPLER_REFERENCE   (NRF_SAADC_REFERENCE_INTERNAL)
/** The acquisition time used for the ADC (currently the lowest to allow for the highest possible baud rate) */
#ifdef NRF_VT_BACKEND_ADC_ACQ_TIME
#define ADC_SAMPLER_ACQ_TIME    (NRF_VT_BACKEND_ADC_ACQ_TIME)
#else
#define ADC_SAMPLER_ACQ_TIME    (NRF_SAADC_ACQTIME_3US)
#endif
/** The burst mode used for the ADC (no burst since no oversample is used) */
#define ADC_SAMPLER_BURST       (NRF_SAADC_BURST_DISABLED)

/** Reference voltage (in milli volts) used by ADC while doing conversion. */
#define ADC_REF_VOLTAGE_IN_MILLIVOLTS   (600)
/** Constant multiplier to compensate for the ADC scaling configuration */
#define ADC_PRE_SCALING_COMPENSATION    (6)
/** Maximum digital value for 10-bit ADC conversion. */
#define ADC_RES_10BIT                   (1024)
#define ADC_RES_10BIT_DIFF              (512)

/** Macro for converting a value in millivolts to a single-ended ADC result */
#define MILLI_VOLTS_TO_ADC_RESULT(MILLI_VOLTS) \
        (nrf_saadc_value_t)((ADC_RES_10BIT * (MILLI_VOLTS)) / (ADC_REF_VOLTAGE_IN_MILLIVOLTS * ADC_PRE_SCALING_COMPENSATION))

/** Macro for converting a value in millivolts to a differential ADC result */
#define MILLI_VOLTS_TO_ADC_RESULT_DIFF(MILLI_VOLTS_DIFF) \
        (nrf_saadc_value_t)((ADC_RES_10BIT_DIFF * (MILLI_VOLTS_DIFF)) / (ADC_REF_VOLTAGE_IN_MILLIVOLTS * ADC_PRE_SCALING_COMPENSATION))


/** Offset of the SAMPLERATE.MODE bit */
#define SAMPLERATE_MODE_OFFSET          (12)
/** Constant value for setting SAMPLERATE.MODE to Task Mode */
#define SAMPLERATE_MODE_TASK            (0)
/** Constant value for setting SAMPLERATE.MODE to Timer Mode */
#define SAMPLERATE_MODE_TIMER           (1)

/** The minimum value allowed for SAMPLERATE.CC */
#define SAMPLERATE_MIN_CC_VALUE         (80)
/** The maximum value allowed for SAMPLERATE.CC */
#define SAMPLERATE_MAX_CC_VALUE         (2047)
/** The minimum sample frequency possible with the internal clock of the SAADC */
#define MIN_SAMPLE_FREQUENCY            (16000000 / SAMPLERATE_MAX_CC_VALUE)
/** The maximum sample frequency possible with the internal clock of the SAADC */
#define MAX_SAMPLE_FREQUENCY            (16000000 / SAMPLERATE_MIN_CC_VALUE)
/** The approximate maximum sample frequency when using 10us acquisition time */
#define MAX_SAMPLE_FREQUENCY_10US_ACQ   (83000)

/** Macro for calculating SAMPLERATE.CC based on a given sample frequency */
#define SAMPLERATE_CC_VALUE(sample_rate) (16000000 / sample_rate)


/*
 * PWM constants / macros
 *****************************************************************************/

/** The count mode used for the PWM */
#define PWM_TRANSMITTER_COUNT_MODE  (NRF_PWM_MODE_UP)
/** The base clock frequency used for the PWM */
#ifdef NRF_VT_BACKEND_PWM_BASE_CLOCK
#define PWM_TRANSMITTER_BASE_CLOCK  (NRF_VT_BACKEND_PWM_BASE_CLOCK)
#else
#define PWM_TRANSMITTER_BASE_CLOCK  (NRF_PWM_CLK_8MHz)
#endif
/** The load mode used for the PWM (only one bit stream is loaded, so it uses the common mode) */
#define PWM_TRANSMITTER_LOAD_MODE   (NRF_PWM_LOAD_COMMON)
/** The step mode used for the PWM */
#define PWM_TRANSMITTER_STEP_MODE   (NRF_PWM_STEP_AUTO)

/** The numerical value of the PWM base clock frequency */
#define PWM_TRANSMITTER_BASE_CLOCK_FREQUENCY (16000000 >> PWM_TRANSMITTER_BASE_CLOCK) 

/** The lowest COUNTERTOP value allowed */
#define MIN_TOP_VALUE (3)
/** The highest COUNTERTOP value allowed */
#define MAX_TOP_VALUE (32767)

/** Macro for converting a baud rate to a COUNTERTOP value */
#define BAUD_RATE_TO_TOP_VALUE(baud_rate) (PWM_TRANSMITTER_BASE_CLOCK_FREQUENCY / baud_rate)

/** The fewest amount of bits that can be transmitted in a sequence */
#define PWM_MIN_BUFFER_SIZE (0)
/** The most amount of bits that can be transmitted in a sequence */
#define PWM_MAX_BUFFER_SIZE (32768)

/** Macro for determining the PWM sequence value for a low bit */
#define PWM_TRANSMITTER_LOW_BIT(_top_value) (_top_value + 1)
/** Macro for determining the PWM sequence value for a high bit */
#define PWM_TRANSMITTER_HIGH_BIT(_top_value) (0)

/** nrfx_pwm flags used when starting a sequence */
#define PWM_TRANSMITTER_SEQUENCE_FLAGS (NRFX_PWM_FLAG_NO_EVT_FINISHED | NRFX_PWM_FLAG_SIGNAL_END_SEQ0 | NRFX_PWM_FLAG_SIGNAL_END_SEQ1)


/*
 * Profiling / debugging
 *****************************************************************************/

#if defined(VT_FRONTEND_DEBUG) || defined(VT_BACKEND_LOG_MEASUREMENTS_ONLY)
#include <stdlib.h>

#define MAX_CHARS_PER_MEASUREMENT (/* Value + sign */ 5)
#define MAX_MEASUREMENTS_PER_PRINT (32)

static char _s_measurement_string[(MAX_CHARS_PER_MEASUREMENT + /* Space/newline */ 1) * MAX_MEASUREMENTS_PER_PRINT + /* \0 */ 1];


static int _saadc_value_to_a(nrf_saadc_value_t N, char *str, int base)
{
    int i = 1;
    int uarg = (int) N;

    char * head = str;
    char * tail;
    char   buf[MAX_CHARS_PER_MEASUREMENT];
    
    tail = &buf[MAX_CHARS_PER_MEASUREMENT - 1];

    if (N < 0)
    {
        *head++ = '-';
        uarg    = -uarg;
    }

    if (uarg)
    {
        div_t r;
        for (i = 0; uarg; ++i)
        {
            r       = div(uarg, base);
            *tail-- = (char)(r.rem + ((9L < r.rem) ? ('A' - 10L) : '0'));
            uarg    = r.quot;
        }
    }
    else
    {
        *tail-- = '0';
    }

    memcpy(head, ++tail, i);

    return (head == str) ? i : i + 1;
}


static size_t _build_measurement_string(nrf_saadc_value_t * p_values, size_t count)
{
    size_t si = 0;
    for (size_t i = 0; i < count; i++)
    {
        si += _saadc_value_to_a(p_values[i], &_s_measurement_string[si], 16);
        _s_measurement_string[si] = ' ';
        si += 1;
    }
    _s_measurement_string[si - 1] = '\n';

    return si;
}


static void print_measurements(nrf_saadc_value_t * p_values, size_t count)
{
    size_t length = _build_measurement_string(p_values, count);
    SEGGER_RTT_Write(0, &_s_measurement_string[0], length);  // Should check for "Up" buffer exhaustion here
}

#endif // VT_BACKEND_LOG_STATS

#ifdef VT_BACKEND_DUMP_PARSED_STREAM
static void print_parsed_word(uint32_t value)
{
    char string[34];
    for (uint8_t i = 0; i < 32; i++)
    {
        string[31 - i] = ((value >> i) & 1) ? '1' : '0';
    }
    string[32] = '\n';
    string[33] = '\0';

    SEGGER_RTT_WriteString(0, string);
}
#endif

#ifdef ADC_SAMPLER_PROFILING
#define PROFILE_SAADC_GPIOTE_TASK_IDX                4
#define PROFILE_SAADC_EVENTS_END_PPI_CHANNEL        (11)
#define PROFILE_SAADC_EVENTS_STARTED_PPI_CHANNEL    (12)

#define PROFILE_PIN_SAADC                           (11)
#define PROFILE_PIN_SAADC_HANDLER                   (12)
#define PROFILE_PIN_ALGORITHM                       (13)
#endif

#ifdef PWM_TRANSMITTER_PROFILING
#define PROFILE_PWM_GPIOTE_TASK_IDX                  5
#define PROFILE_PWM_EVENTS_SEQSTARTED_0_PPI_CHANNEL (13)
#define PROFILE_PWM_EVENTS_SEQSTARTED_1_PPI_CHANNEL (14)
#define PROFILE_PWM_EVENTS_SEQEND_0_PPI_CHANNEL     (15)
#define PROFILE_PWM_EVENTS_SEQEND_1_PPI_CHANNEL     (16)

#define PROFILE_PIN_PWM                             (16)
#define PROFILE_PIN_TX_FUNCTION                     (17)
#endif


/*
 * Sampling constants / configuration
 *****************************************************************************/

/** SAADC channel used */
#if defined(NRF_VT_BACKEND_SAADC_CHANNEL)
#define SAADC_CHANNEL (NRF_VT_BACKEND_SAADC_CHANNEL)

#if SAADC_CHANNEL < 0 || SAADC_CHANNEL > NRF_SAADC_CHANNEL_COUNT
#error "Invalid SAADC channel chosen"
#endif // Option validation

#else
#define SAADC_CHANNEL (0)
#endif

/** Samples per bit cycle */
#if defined(NRF_VT_BACKEND_SAMPLES_PER_BIT)
#define SAMPLES_PER_BIT (NRF_VT_BACKEND_SAMPLES_PER_BIT)
#endif

/** Samples per sample buffer (i.e. buffer size) */
#if defined(NRF_VT_BACKEND_RX_BUFFER_SIZE)
#define SAMPLE_BUFFER_SAMPLES (NRF_VT_BACKEND_RX_BUFFER_SIZE)
#else
#define SAMPLE_BUFFER_SAMPLES (1024)
#endif

/** Minimum permissible samples made per bit */
#define MIN_SAMPLES_PER_BIT (3)
/** Maximum permissible samples made per bit */
#define MAX_SAMPLES_PER_BIT (7)
/** Permissible steps between MIN_SAMPLES_PER_BIT and MAX_SAMPLES_PER_BIT */
#define STEP_SAMPLES_PER_BIT (2)

/** The fewest amount of consecutive bit (samples) that can be considered a single bit */
#define MIN_CONSECUTIVE_BITS(SAMPLES_PER_BIT) (SAMPLES_PER_BIT - 1)

/** Macro used for converting an ADC value to a bit */
#define MEASUREMENT_BIT_VALUE(measurement, threshold) ((measurement < threshold) ? 0 : 1)

/** The index of the sample buffers at which the measurements start */
#define SAMPLE_BUFFER_MEASUREMENTS_START (0)
/** The total size of each sample buffer */
#define SAMPLE_BUFFER_TOTAL_SIZE (SAMPLE_BUFFER_SAMPLES)

/** The amount of buffers reserved for direct writing by the ADC */ 
#define CORE_BUFFER_COUNT (2)

/** The amount of extra buffers used for storing samples */
#if defined(NRF_VT_BACKEND_COPY_BUFFER_COUNT)
#define COPY_BUFFER_COUNT (NRF_VT_BACKEND_COPY_BUFFER_COUNT)
#else
#define COPY_BUFFER_COUNT (2)
#endif

/** The total amount of buffers used */
#define SAMPLE_BUFFER_COUNT (CORE_BUFFER_COUNT + COPY_BUFFER_COUNT)

/** The PPI channel used to trigger TASK_START at the end of a buffer */
#if defined(NRF_VT_BACKEND_ADC_START_PPI_CHANNEL)
#define ADC_SAMPLER_START_PPI_CHANNEL (NRF_VT_BACKEND_ADC_START_PPI_CHANNEL)
#else
#define ADC_SAMPLER_START_PPI_CHANNEL (10)
#endif

#if ADC_SAMPLER_START_PPI_CHANNEL < 0 || ADC_SAMPLER_START_PPI_CHANNEL > 19
#error "Invalid PPI channel selected. Please select a channel between 0 and 19 (inclusive)."
#endif

/*
 * Transmitter configuration
 *****************************************************************************/

/** The PWM module used */
#ifdef NRF_VT_BACKEND_PWM_MODULE
#define PWM_TRANSMITTER_PWM_MODULE NRF_VT_BACKEND_PWM_MODULE

#if PWM_TRANSMITTER_PWM_MODULE < 0 || PWM_TRANSMITTER_PWM_MODULE > NRF_PWM_CHANNEL_COUNT
#error "Invalid PWM module selected"
#endif // Validation

#else
#define PWM_TRANSMITTER_PWM_MODULE 0  // PWM0
#endif

/** The size of the TX buffer - uses 2B per bit */
#ifdef NRF_VT_BACKEND_TX_BUFFER_SIZE
#define PWM_TRANSMITTER_TX_BUFFER_SIZE      (NRF_VT_BACKEND_TX_BUFFER_SIZE)
#else 
#define PWM_TRANSMITTER_TX_BUFFER_SIZE      (4096)
#endif

/** Maximum amount of start bits for a frame */
#define PWM_TRANSMITTER_MAX_START_BITS      (8)
/** Maximum amount of stop bits for a frame */
#define PWM_TRANSMITTER_MAX_STOP_BITS       (8)
/** Maximum amount of idle time after a frame */
#define PWM_TRANSMITTER_MAX_END_DELAY       (8)
/** Maximum amount of payload bits in a frame */
#define PWM_TRANSMITTER_MAX_PAYLOAD_BITS    (32)
/** Maximum frame length in bits */
#define PWM_TRANSMITTER_MAX_FRAME_BITS      (PWM_TRANSMITTER_MAX_START_BITS + PWM_TRANSMITTER_MAX_STOP_BITS \
                                                + PWM_TRANSMITTER_MAX_END_DELAY + PWM_TRANSMITTER_MAX_PAYLOAD_BITS)

/** Amount of frame casts in the transmitter backend */
#if defined(NRF_VT_BACKEND_FRAME_CAST_COUNT)
#define PWM_TRANSMITTER_FRAME_CAST_COUNT (NRF_VT_BACKEND_FRAME_CAST_COUNT)

#if PWM_TRANSMITTER_FRAME_CAST_COUNT < 1
#error "Invalid frame cast count. Currently there must be at least one frame cast."
#endif // Validation

#else
#define PWM_TRANSMITTER_FRAME_CAST_COUNT    (2)
#endif


/*****************************************************************************
 * Externally defined values
 *****************************************************************************/

/* These values / functions are defined by the front-end in use */

#ifdef VT_FRONTEND_DEBUG
uint32_t vt_frontend_parse_next_bit(uint_fast8_t bit);
#else
bool vt_frontend_parse_next_bit(uint_fast8_t bit);
#endif
void vt_frontend_overrun_handler(void);
void vt_frontend_tx_done_handler(void);


/*****************************************************************************
 * Static variables
 *****************************************************************************/

/*
 * Common variables
 *****************************************************************************/

static uint8_t m_rx_pin_p = NRF_VT_PIN_DISCONNECTED;
static uint8_t m_rx_pin_n = NRF_VT_PIN_DISCONNECTED;
static uint8_t m_tx_pin = NRF_VT_PIN_DISCONNECTED;


/*
 * Sampler variables
 *****************************************************************************/

/** Queue node type used in the buffer FIFO */
typedef struct {
    nrf_saadc_value_t * volatile p_buffer;
    bool volatile invalidate_after;
}
sample_buffer_node_t;

/** The buffers used for storing ADC samples */
static nrf_saadc_value_t        m_rx_buffers[SAMPLE_BUFFER_COUNT][SAMPLE_BUFFER_TOTAL_SIZE];
/** Helper structure used for implementing a FIFO queue of sample buffers */
static sample_buffer_node_t     m_rx_nodes[SAMPLE_BUFFER_COUNT];

/** Read index for the RX nodes */
static volatile uint_fast8_t    m_rx_rd;
/** Write index for the RX nodes (i.e. the index of the node pointing to the primary ADC buffer) */
static volatile uint_fast8_t    m_rx_wr;
/** Next write index for the RX nodes (i.e. the index of the node pointing to the secondary ADC buffer) */
static volatile uint_fast8_t    m_rx_wr_next;

/** Number of samples per bit */
static uint_fast8_t             m_samples_per_bit = MIN_SAMPLES_PER_BIT;

/** Variable indicating whether sample processing is currently running */
static volatile bool            m_processing_data = false;

/** Current threshold value used to distinguish 0 and 1 bits */
static nrf_saadc_value_t        m_threshold_value = INT16_MIN;


/*
 * Transmitter variables
 *****************************************************************************/

/** PWM instance used for the transmitter */
static nrfx_pwm_t               m_pwm = NRFX_PWM_INSTANCE(PWM_TRANSMITTER_PWM_MODULE);

/** COUNTERTOP value used to write 0/1 bits */
static uint32_t                 m_top_value = 0;

/** TX buffer used for storing the PWM sequence being transmitted */
static nrf_pwm_values_common_t  m_tx_buffer[PWM_TRANSMITTER_TX_BUFFER_SIZE];
/** Length of the TX buffer contents */
static uint32_t                 m_tx_data_length = 0;

/** Is a PWM sequence currently executing? */
static volatile bool            m_sequence_running = false;
/** The sequence currently being executed */
static nrf_pwm_sequence_t       m_current_sequence;

/** Lookup table for constructing PWM sequences based on given bytes */
static nrf_pwm_values_common_t  m_lookup_table[UINT8_MAX + 1][8];

/** 'Casts' for more efficiently creating PWM sequences */
static nrf_pwm_values_common_t  m_frame_casts[PWM_TRANSMITTER_FRAME_CAST_COUNT][PWM_TRANSMITTER_MAX_FRAME_BITS];
/** Payload length for the frame formats for each cast */
static uint8_t                  m_frame_payload_lengths[PWM_TRANSMITTER_FRAME_CAST_COUNT];


/*****************************************************************************
 * Macros / static functions
 *****************************************************************************/

/*
 * RX implementation
 *****************************************************************************/

/** Macro for getting the next index in a circular buffer */
#define CIRCULAR_BUFFER_PREV_INDEX(current, size) ((current == 0) ? size - 1 : current - 1)
/** Macro for getting the previous index in a circular buffer */
#define CIRCULAR_BUFFER_NEXT_INDEX(current, size) ((current == (size - 1)) ? 0 : current + 1) 


/**@brief Function for getting the node pointing to the next readable sample buffer */
static inline sample_buffer_node_t * __attribute__((always_inline)) read_node_get(void)
{
    if (m_rx_rd == m_rx_wr)
    {
        return NULL;
    }
    return &m_rx_nodes[m_rx_rd];
}


/**@brief Function for moving the read pointer to the next position in the sample buffer queue */
static inline void __attribute__((always_inline)) read_pointer_move(void)
{
    m_rx_nodes[m_rx_rd].invalidate_after = false;
    m_rx_rd = CIRCULAR_BUFFER_NEXT_INDEX(m_rx_rd, SAMPLE_BUFFER_COUNT);
}


/**@brief Function for getting the next ADC primary buffer */
static inline nrf_saadc_value_t * volatile __attribute__((always_inline)) write_buffer_get(void)
{
    return m_rx_nodes[m_rx_wr].p_buffer;
}


/**@brief Function for getting the next ADC secondary buffer */
static inline nrf_saadc_value_t * volatile __attribute__((always_inline)) write_buffer_next_get(void)
{
    return m_rx_nodes[m_rx_wr_next].p_buffer;
}


/**@brief Function for moving the primary and secondary buffer pointers in the sample buffer queue */
static inline void __attribute__((always_inline)) write_pointers_move(void)
{
    uint_fast8_t rx_wr_next_next = CIRCULAR_BUFFER_NEXT_INDEX(m_rx_wr_next, SAMPLE_BUFFER_COUNT);
    
    // Detect if the buffers are exhausted
    if (rx_wr_next_next == m_rx_rd)
    {
        m_rx_nodes[CIRCULAR_BUFFER_PREV_INDEX(m_rx_wr, SAMPLE_BUFFER_COUNT)].invalidate_after = true;
        nrf_saadc_value_t * volatile p_tmp = m_rx_nodes[m_rx_wr].p_buffer;
        m_rx_nodes[m_rx_wr].p_buffer = m_rx_nodes[m_rx_wr_next].p_buffer;
        m_rx_nodes[m_rx_wr_next].p_buffer = p_tmp;
    }
    else
    {
        m_rx_wr = m_rx_wr_next;
        m_rx_wr_next = rx_wr_next_next;
    }
/*
#ifdef VT_BACKEND_LOG_STATS
    SEGGER_RTT_printf(0, "W m_rx_wr: %d, p_buffer: 0x%08X; m_rw_wr_next: %d, p_buffer: 0x%08X\n",
                         m_rx_wr, m_rx_nodes[m_rx_wr].p_buffer, m_rx_wr_next, m_rx_nodes[m_rx_wr_next].p_buffer);
#endif
*/
}


static inline void __attribute__((always_inline)) setup_next_write_buffer(void)
{
    // At this point, the buffer pointed to by the write pointer is already filled, so the pointer can be moved
    write_pointers_move();
    nrfx_saadc_buffer_convert(write_buffer_next_get(), SAMPLE_BUFFER_SAMPLES);
}


static inline void __attribute__((always_inline)) setup_initial_write_buffers(void)
{
    APP_ERROR_CHECK(nrfx_saadc_buffer_convert(write_buffer_get(), SAMPLE_BUFFER_SAMPLES));
    APP_ERROR_CHECK(nrfx_saadc_buffer_convert(write_buffer_next_get(), SAMPLE_BUFFER_SAMPLES));  
}


/**@brief Function for initializing the sample buffer queue */
static void sample_buffers_init(void)
{
    for (int i = 0; i < SAMPLE_BUFFER_COUNT; i++)
    {
        m_rx_nodes[i] = (sample_buffer_node_t) {
            .p_buffer = &m_rx_buffers[i][0],
            .invalidate_after = false
        };
    }
    m_rx_rd = 0;
    m_rx_wr = 0;
    m_rx_wr_next = CIRCULAR_BUFFER_NEXT_INDEX(m_rx_wr, SAMPLE_BUFFER_COUNT);
}


/**@brief Function for converting a pin number to a nrf_saadc_input_t
 */
static nrf_saadc_input_t pin_to_saadc_input(uint8_t pin)
{
    switch (pin)
    {
        case 2: return NRF_SAADC_INPUT_AIN0;
        case 3: return NRF_SAADC_INPUT_AIN1;
        case 4: return NRF_SAADC_INPUT_AIN2;
        case 5: return NRF_SAADC_INPUT_AIN3;
        case 28: return NRF_SAADC_INPUT_AIN4;
        case 29: return NRF_SAADC_INPUT_AIN5;
        case 30: return NRF_SAADC_INPUT_AIN6;
        case 31: return NRF_SAADC_INPUT_AIN7;
        default: return NRF_SAADC_INPUT_DISABLED; 
    }
}


/**@brief Function for calculating the least possible samples per bit for a given baud_rate
 */
static uint_fast8_t least_samples_per_bit(uint32_t baud_rate)
{
    uint32_t real_rate;
    for (uint_fast8_t i = MIN_SAMPLES_PER_BIT; i <= MAX_SAMPLES_PER_BIT; i += STEP_SAMPLES_PER_BIT)
    {
        real_rate = baud_rate * i;
        if (MIN_SAMPLE_FREQUENCY < real_rate && real_rate < MAX_SAMPLE_FREQUENCY)
        {
            return i;
        }
    }
    return 0;
}


/**@brief Function for processing the samples generated by the ADC
 */
static inline __attribute__((always_inline)) void samples_process(nrf_saadc_value_t volatile * p_values, bool invalidate_after)
{
#ifdef ADC_SAMPLER_PROFILING
    nrf_gpio_pin_set(PROFILE_PIN_ALGORITHM);
#endif

#ifdef VT_BACKEND_LOG_MEASUREMENTS_ONLY
    for (size_t i = SAMPLE_BUFFER_MEASUREMENTS_START; i < SAMPLE_BUFFER_TOTAL_SIZE; i += MAX_MEASUREMENTS_PER_PRINT)
    {
        print_measurements(&p_values[i], MIN(SAMPLE_BUFFER_TOTAL_SIZE - i, MAX_MEASUREMENTS_PER_PRINT));
    }
#else

#ifdef VT_BACKEND_DUMP_PARSED_STREAM
    static uint32_t parsed_word = 0;
    static int parsed_word_pos = 31;
#endif

#ifdef VT_FRONTEND_DEBUG
    static uint_fast32_t sample_index;
#else
    uint_fast32_t sample_index;
#endif
    sample_index = 0;
    nrf_saadc_value_t threshold_value = m_threshold_value;

    uint_fast8_t current;
    static uint_fast8_t last;
    static uint_fast8_t consecutive = 0;

#ifdef VT_FRONTEND_DEBUG
    uint32_t res;
#endif

    // Initialize the "last" variable before the upcoming loop only if there is not a previous "last" value
    // that should be considered
    if (consecutive == 0)
    {
        last = 2;  // Sentinel value so that current != last
    }
	
    while (sample_index < SAMPLE_BUFFER_TOTAL_SIZE)
    {
        do
        {
            current = MEASUREMENT_BIT_VALUE(p_values[sample_index], threshold_value);
        
            if (current != last)
            {
                if (consecutive >= MIN_CONSECUTIVE_BITS(m_samples_per_bit))
                {
                    // In this case we should accept the bit as the value of 'last', but
                    // we should not move past the current sample index
                    consecutive = 0;
                    break;
                }
                else
                {
                    consecutive = 1;
                }
            }
            else
            {
                consecutive++;
            }

            last = current;
            sample_index++;
        
            if (consecutive >= m_samples_per_bit)
            {
                consecutive = 0;
                break;
            }
        }
        while (sample_index < SAMPLE_BUFFER_TOTAL_SIZE);

#ifdef VT_BACKEND_DUMP_PARSED_STREAM
        if (consecutive == 0)
        {
            parsed_word |= (last << parsed_word_pos);
            parsed_word_pos--;
            if (parsed_word_pos < 0)
            {
                print_parsed_word(parsed_word);
                parsed_word = 0;
                parsed_word_pos = 31;
            }
        }
#else
        // consecutive == 0 is what should indicate that a bit was parsed.
        if (consecutive == 0)
        {
#ifdef VT_FRONTEND_DEBUG
            if ((res = vt_frontend_parse_next_bit(last)) != 0)
            {
                if (res != 1)
                {
                    if (sample_index < MAX_MEASUREMENTS_PER_PRINT)
                    {
                        SEGGER_RTT_WriteString(0, "End of last buffer:\n");
                        print_measurements(&m_rx_nodes[CIRCULAR_BUFFER_PREV_INDEX(m_rx_rd, SAMPLE_BUFFER_COUNT)]
                                                .p_buffer[SAMPLE_BUFFER_TOTAL_SIZE - MAX_MEASUREMENTS_PER_PRINT - 1], MAX_MEASUREMENTS_PER_PRINT);
                    }

                    int lower = MAX((int)(sample_index) - MAX_MEASUREMENTS_PER_PRINT / 2, SAMPLE_BUFFER_MEASUREMENTS_START);
                    int upper = MIN(lower + MAX_MEASUREMENTS_PER_PRINT, SAMPLE_BUFFER_TOTAL_SIZE);

                    print_measurements(&p_values[lower], sample_index - lower);
                
                    if (upper - sample_index > 0)
                    {
                        SEGGER_RTT_WriteString(0, "| ");
                        print_measurements(&p_values[sample_index], upper - sample_index);
                    }
                    else
                    {
                        SEGGER_RTT_WriteString(0, "|\n");
                    }
                }
#else
            if (!vt_frontend_parse_next_bit(last)) // TODO: optimize
            {
#endif // VT_FRONTEND_DEBUG

                if (_VT_START_BIT_VALUE)
                {
                    while (p_values[sample_index] < threshold_value && sample_index < SAMPLE_BUFFER_TOTAL_SIZE)
                    {
                        sample_index++;
                    }
                }
                else
                {
                    while (p_values[sample_index] >= threshold_value && sample_index < SAMPLE_BUFFER_TOTAL_SIZE)
                    {
                        sample_index++;
                    }
                }

                if (sample_index < SAMPLE_BUFFER_TOTAL_SIZE)
                {
                    last = _VT_START_BIT_VALUE;
                    consecutive = 1;
                    sample_index++;
                }

                // else: consecutive should still be equal to 0, and the surrounding 
                //       while-loop should exit since sample_index == SAMPLE_BUFFER_TOTAL_SIZE
            }
        }
#endif // VT_BACKEND_DUMP_PARSED_STREAM
    }

    if (invalidate_after)
    {
        vt_frontend_overrun_handler(); // This will reset the frontend parser
        consecutive = 0;  // This should effectively reset this function
    }
#endif // VT_BACKEND_LOG_MEASUREMENTS_ONLY

#ifdef ADC_SAMPLER_PROFILING
    nrf_gpio_pin_clear(PROFILE_PIN_ALGORITHM);
#endif
}


/**@brief Event handler function for processing ADC samples */
static void app_sched_process_handler(void * p_event_data, uint16_t event_size)
{
    UNUSED_PARAMETER(p_event_data);
    UNUSED_PARAMETER(event_size);

    sample_buffer_node_t * p_node;

    m_processing_data = true;

#ifdef VT_BACKEND_LOG_STATS
    SEGGER_RTT_WriteString(0, "Processing...\n");
#endif
    while ((p_node = read_node_get()) != NULL)
    {
#ifdef VT_BACKEND_LOG_STATS
        SEGGER_RTT_printf(0, "P rx_rd: %d, p_buffer: 0x%08X, invalidate_after: %d\n",
                             m_rx_rd, p_node->p_buffer, p_node->invalidate_after);
#endif
        samples_process(p_node->p_buffer, p_node->invalidate_after);
        read_pointer_move();
    }
    
    m_processing_data = false;
}


/**@brief Function for handling events from nrfx_saadc */
static void saadc_event_handler(nrfx_saadc_evt_t const * p_event)
{
#ifdef ADC_SAMPLER_PROFILING
    nrf_gpio_pin_set(PROFILE_PIN_SAADC_HANDLER);
#endif
    switch (p_event->type)
    {
        case NRFX_SAADC_EVT_DONE:
        {
            setup_next_write_buffer();
            app_sched_event_put(NULL, 0, app_sched_process_handler);
            break;
        }

        case NRFX_SAADC_EVT_CALIBRATEDONE:
        {
            setup_initial_write_buffers();
            while (nrfx_saadc_sample() == NRFX_ERROR_INVALID_STATE) {}
            break;
        }

        case NRFX_SAADC_EVT_LIMIT:
        {
            break;
        }

        default:
        {
            break;
        }
    }
#ifdef ADC_SAMPLER_PROFILING
    nrf_gpio_pin_clear(PROFILE_PIN_SAADC_HANDLER);
#endif
}


/**@brief Function for setting up and enabling the PPI channel for triggering TASK_START after EVENT_END */
static bool adc_sampler_ppi_channel_init(void)
{
    if (nrf_ppi_channel_enable_get(ADC_SAMPLER_START_PPI_CHANNEL) == NRF_PPI_CHANNEL_ENABLED)
    {
        return false;
    }
    nrf_ppi_channel_endpoint_setup(ADC_SAMPLER_START_PPI_CHANNEL,
        nrf_saadc_event_address_get(NRF_SAADC_EVENT_END),
        nrf_saadc_task_address_get(NRF_SAADC_TASK_START));
    nrf_ppi_channel_enable(ADC_SAMPLER_START_PPI_CHANNEL);
    return true;
}


/**@brief Function for disabling the PPI channel used for triggering TASK_START after EVENT_END */
static void adc_sampler_ppi_channel_uninit(void)
{
    nrf_ppi_channel_disable(ADC_SAMPLER_START_PPI_CHANNEL);
}


/**@brief Function for initializing the ADC */
static uint32_t adc_sampler_init(uint32_t source_baud_rate, int16_t threshold_milli_volts, uint8_t saadc_irq_priority, bool differential)
{
    nrfx_err_t nrfx_err;
  
    if (!INTERRUPT_PRIORITY_IS_VALID(saadc_irq_priority))
    {
        return NRF_ERROR_INVALID_PARAM;
    }

#ifdef SAMPLES_PER_BIT
    uint_fast8_t samples_per_bit = SAMPLES_PER_BIT;
#else
    uint_fast8_t samples_per_bit = least_samples_per_bit(source_baud_rate);
#endif

    if (samples_per_bit == 0)
    {
        return NRF_ERROR_INVALID_PARAM;
    }
    uint32_t sample_rate = source_baud_rate * samples_per_bit;
    
    // While the below PAN is present, it only prevents EVENT_STARTED from firing.
    // Luckily we do not need EVENT_STARTED for the receiver to function correctly.
#if defined(NRF52_PAN_150) && 0 
    if (sample_rate > MAX_SAMPLE_FREQUENCY_10US_ACQ)
    {
        return NRF_ERROR_INVALID_PARAM;
    }
#endif // NRF52_PAN_150

    if (!adc_sampler_ppi_channel_init())
    {
        // The PPI channel is already in use
        return NRF_ERROR_BUSY;
    }

    nrfx_saadc_config_t saadc_config = NRFX_SAADC_DEFAULT_CONFIG;
    saadc_config.resolution = ADC_SAMPLER_RESOLUTION;
    saadc_config.oversample = ADC_SAMPLER_OVERSAMPLE;
    saadc_config.low_power_mode = ADC_SAMPLER_LP_MODE;
    saadc_config.interrupt_priority = saadc_irq_priority;
    nrfx_err = nrfx_saadc_init(&saadc_config, saadc_event_handler);

    if (nrfx_err != NRFX_SUCCESS)
    {
        adc_sampler_ppi_channel_uninit();

        NRF_LOG_ERROR("SAADC initialization failed with error: %s", NRFX_LOG_ERROR_STRING_GET(nrfx_err));
        return NRF_ERROR_INVALID_STATE;
    }

    // For some reason, setting Continuous Mode is not a part of the driver/HAL...
    uint32_t cc_value = SAMPLERATE_CC_VALUE(sample_rate);
    NRF_SAADC->SAMPLERATE = (SAMPLERATE_MODE_TIMER << SAMPLERATE_MODE_OFFSET) | cc_value;

    m_samples_per_bit = samples_per_bit;
    
    if (!differential)
    {
        m_threshold_value = MILLI_VOLTS_TO_ADC_RESULT(threshold_milli_volts);
    }
    else
    {
        m_threshold_value = MILLI_VOLTS_TO_ADC_RESULT_DIFF(threshold_milli_volts);
    }   

    sample_buffers_init();
    

    return NRF_SUCCESS;
}


/**@brief Function for uninitializing the ADC sampler */
static void adc_sampler_uninit()
{
    adc_sampler_ppi_channel_uninit();
    nrfx_saadc_uninit();

    NRF_SAADC->SAMPLERATE = 0;
    m_threshold_value = INT16_MIN;
}


/**@brief Function for initializing an ADC channel for a given pin */
static uint32_t adc_sampler_pin_init(uint8_t pin_p, uint8_t pin_n)
{
    nrfx_err_t nrfx_err;
    nrf_saadc_channel_config_t channel_config;
    
    nrf_saadc_input_t input_p;
    nrf_saadc_input_t input_n;

    input_p = pin_to_saadc_input(pin_p);
    if (input_p == NRF_SAADC_INPUT_DISABLED && pin_p != NRF_VT_PIN_DISCONNECTED)
    {
        return NRF_ERROR_INVALID_PARAM;
    }

    input_n = pin_to_saadc_input(pin_n);
    if (input_n == NRF_SAADC_INPUT_DISABLED && pin_n != NRF_VT_PIN_DISCONNECTED)
    {
        return NRF_ERROR_INVALID_PARAM;
    }
    
    if (pin_n == NRF_VT_PIN_DISCONNECTED)
    {
        channel_config.mode = NRF_SAADC_MODE_SINGLE_ENDED;
    }
    else
    {
        channel_config.mode = NRF_SAADC_MODE_DIFFERENTIAL;
    }
    channel_config.acq_time = ADC_SAMPLER_ACQ_TIME;
    channel_config.burst = ADC_SAMPLER_BURST;
    channel_config.gain = ADC_SAMPLER_GAIN;
    channel_config.pin_p = input_p;
    channel_config.pin_n = input_n;
    channel_config.resistor_p = ADC_SAMPLER_RESISTOR_P;
    channel_config.resistor_n = ADC_SAMPLER_RESISTOR_N;
    channel_config.reference = ADC_SAMPLER_REFERENCE;

    nrfx_err = nrfx_saadc_channel_init(SAADC_CHANNEL, &channel_config);
    if (nrfx_err != NRFX_SUCCESS)
    {
        NRF_LOG_ERROR("SAADC channel initialization failed with error: %s", NRFX_LOG_ERROR_STRING_GET(nrfx_err));
        if (nrfx_err == NRFX_ERROR_BUSY)
        {
            return NRF_ERROR_BUSY;
        }
        else
        {
            return NRF_ERROR_INVALID_STATE;
        }
    }

    return NRF_SUCCESS;
}


/**@brief Function for uninitializing the ADC channel in use */
static void adc_sampler_pin_uninit(void)
{
    nrfx_saadc_channel_uninit(SAADC_CHANNEL);
}


/*
 * TX implementation
 *****************************************************************************/

/**@brief Construct the PWM sequence value lookup table */
static void lookup_table_init(uint32_t top_value, bool lsbit_first)
{
    nrf_pwm_values_common_t low_bit = PWM_TRANSMITTER_LOW_BIT(top_value);
    nrf_pwm_values_common_t high_bit = PWM_TRANSMITTER_HIGH_BIT(top_value);
    if (lsbit_first)
    {
        for (uint32_t i = 0; i <= 0xff; i++)
        {
            for (uint32_t j = 0; j  <= 7; j++)
            {
                m_lookup_table[i][j] = IS_SET(i, j) ? high_bit : low_bit;
            }
        }
    }
    else
    {
        for (uint32_t i = 0; i <= 0xff; i++)
        {
            for (int32_t j = 7; j  >= 0; j--)
            {
                m_lookup_table[i][j] = IS_SET(i, j) ? high_bit : low_bit;
            }
        }
    }
}


/**@brief Macro for extending the lsbit to the least significant 8 bits */
#define EXTEND_LSBIT_TO_8(bit) ((0x100 - bit) & 0xff)


/**@brief Function for initializing a frame cast */
static void frame_cast_init(uint8_t cast_no, uint32_t payload_size)
{
    nrf_saadc_value_t * p_frame_cast = &m_frame_casts[cast_no][0];
    memcpy_fast(p_frame_cast, &m_lookup_table[EXTEND_LSBIT_TO_8(_VT_START_BIT_VALUE)][0], _VT_START_BITS * sizeof(nrf_pwm_values_common_t));
    p_frame_cast += _VT_START_BITS + payload_size;
    memcpy_fast(p_frame_cast, &m_lookup_table[EXTEND_LSBIT_TO_8(_VT_STOP_BIT_VALUE)][0], _VT_STOP_BITS * sizeof(nrf_pwm_values_common_t));
    p_frame_cast += _VT_STOP_BITS;
    memcpy_fast(p_frame_cast, &m_lookup_table[EXTEND_LSBIT_TO_8(_VT_DEFAULT_LINE_VALUE)][0], _VT_END_DELAY * sizeof(nrf_pwm_values_common_t));
    m_frame_payload_lengths[cast_no] = payload_size;
}


/**@brief Function for witing a payload to a frame cast */
static inline void __attribute__((always_inline)) tx_buffer_write_cast_payload(uint8_t cast_no, const uint8_t const * p_data,
                                                                               nrf_pwm_values_common_t * p_buffer, uint32_t word_length)
{
    nrf_pwm_values_common_t * p_lookup;
    uint32_t quotient = word_length >> 3;
    uint32_t remainder = word_length - (quotient << 3);
    nrf_saadc_value_t * p_frame_cast = &m_frame_casts[cast_no][_VT_START_BITS];
    
#if _VT_LSBIT_FIRST
    if (remainder > 0)
    {
        p_lookup = &m_lookup_table[p_data[quotient]][0]; 
        memcpy_fast(p_frame_cast, p_lookup, remainder * sizeof(nrf_pwm_values_common_t));
        p_frame_cast += remainder;
    }
    for (int i = quotient - 1; i >= 0; i--)
    {
        p_lookup = &m_lookup_table[p_data[i]][0];
        memcpy_fast(p_frame_cast, p_lookup, 8 * sizeof(nrf_pwm_values_common_t));
        p_frame_cast += 8;
    }

#else
    for (uint32_t i = 0; i < quotient; i++)
    {
        p_lookup = &m_lookup_table[p_data[i]][0];
        memcpy_fast(p_frame_cast, p_lookup, 8 * sizeof(nrf_pwm_values_common_t));
        p_frame_cast += 8;
    }

    if (remainder > 0)
    {
        p_lookup = &m_lookup_table[p_data[quotient]][8 - remainder]; 
        memcpy_fast(p_frame_cast, p_lookup, remainder * sizeof(nrf_pwm_values_common_t));
        p_frame_cast += remainder;
    }

#endif
    memcpy_fast(p_buffer, &m_frame_casts[cast_no][0],
                (_VT_START_BITS + word_length + _VT_STOP_BITS + _VT_END_DELAY) * sizeof(nrf_pwm_values_common_t));
}


/**@brief Function for writing a frame directly to the PWM buffer (slower than using a frame cast) */
static void tx_buffer_write_frame(const uint8_t const * p_data, nrf_pwm_values_common_t * p_buffer,
                                  uint32_t word_length)
{
    nrf_pwm_values_common_t * p_lookup;

    memcpy_fast(p_buffer, &m_lookup_table[EXTEND_LSBIT_TO_8(_VT_START_BIT_VALUE)][0], _VT_START_BITS * sizeof(nrf_pwm_values_common_t));
    p_buffer += _VT_START_BITS;

    uint32_t quotient = word_length >> 3;
    uint32_t remainder = word_length - (quotient << 3);
    for (uint32_t i = 0; i < quotient; i++)
    {
        p_lookup = &m_lookup_table[p_data[i]][0];
        memcpy_fast(p_buffer, p_lookup, 8 * sizeof(nrf_pwm_values_common_t));
        p_buffer += 8;
    }
    if (remainder > 0)
    {
        p_lookup = &m_lookup_table[p_data[quotient]][8 - remainder]; 
        memcpy_fast(p_buffer, p_lookup, remainder * sizeof(nrf_pwm_values_common_t));
        p_buffer += remainder; 
    }

    memcpy_fast(p_buffer, &m_lookup_table[EXTEND_LSBIT_TO_8(_VT_STOP_BIT_VALUE)][0], _VT_STOP_BITS * sizeof(nrf_pwm_values_common_t));
    p_buffer += _VT_STOP_BITS;
   
    memcpy_fast(p_buffer, &m_lookup_table[EXTEND_LSBIT_TO_8(_VT_DEFAULT_LINE_VALUE)][0], _VT_END_DELAY * sizeof(nrf_pwm_values_common_t));
}


/**@brief Function for handling events from the nrfx_pwm module */
static void pwm_event_handler(nrfx_pwm_evt_type_t event)
{
    switch (event)
    {
        case NRFX_PWM_EVT_FINISHED:
        {
            break;
        }
        case NRFX_PWM_EVT_END_SEQ0:
        case NRFX_PWM_EVT_END_SEQ1:
        {
            m_tx_data_length = 0;
            m_sequence_running = false;
            vt_frontend_tx_done_handler();  // FIXME: should this really be called from interrupt?
            break;
        }
        case NRFX_PWM_EVT_STOPPED:
        {
            break;
        }
    }
}


/**@brief Function for initializing the PWM */
static uint32_t pwm_transmitter_init(uint32_t baud_rate, uint8_t pin, uint8_t pwm_irq_priority)
{
    if (pin > 31)
    {
        return NRF_ERROR_INVALID_PARAM;
    }
    
    if (!INTERRUPT_PRIORITY_IS_VALID(pwm_irq_priority))
    {
        return NRF_ERROR_INVALID_PARAM;
    }

    uint32_t top_value = BAUD_RATE_TO_TOP_VALUE(baud_rate);
    if (top_value >= UINT16_MAX)
    {
        return NRF_ERROR_INVALID_PARAM;
    }
    
    nrfx_err_t nrfx_err;

    nrfx_pwm_config_t config = NRFX_PWM_DEFAULT_CONFIG;
    config.output_pins[0] = pin;
    for (uint8_t i = 1; i < NRF_PWM_CHANNEL_COUNT; i++)
    {
        config.output_pins[i] = NRFX_PWM_PIN_NOT_USED;
    }
    config.irq_priority = pwm_irq_priority;
    config.base_clock = (nrf_pwm_clk_t)PWM_TRANSMITTER_BASE_CLOCK;
    config.count_mode = (nrf_pwm_mode_t)PWM_TRANSMITTER_COUNT_MODE;
    config.top_value = (uint16_t)top_value;
    config.load_mode = (nrf_pwm_dec_load_t)PWM_TRANSMITTER_LOAD_MODE;
    config.step_mode = (nrf_pwm_dec_step_t)PWM_TRANSMITTER_STEP_MODE;
    
    nrfx_err = nrfx_pwm_init(&m_pwm, &config, pwm_event_handler);
    if (nrfx_err != NRFX_SUCCESS)
    {
        return NRF_ERROR_INVALID_STATE;
    }

    nrf_gpio_pin_dir_set(pin, NRF_GPIO_PIN_DIR_OUTPUT);
    if (_VT_DEFAULT_LINE_VALUE)
    {
        nrf_gpio_pin_set(pin);
    }
    else
    {
        nrf_gpio_pin_clear(pin);
    }

    m_top_value = top_value;

    lookup_table_init(m_top_value, _VT_LSBIT_FIRST);

    return NRF_SUCCESS;
}


/**@brief Function for uninitializing the PWM */
static void pwm_transmitter_uninit(uint8_t pin)
{
    nrfx_pwm_uninit(&m_pwm);
    nrf_gpio_cfg_default(pin);
    
    m_top_value = 0;
    m_tx_data_length = 0;
    m_sequence_running = false;
}


/**@brief Function for starting transmission of the current PWM buffer */
static void pwm_transmitter_start_sequence(void)
{
    m_current_sequence.values = (nrf_pwm_values_t) {.p_common = &m_tx_buffer[0]};
    m_current_sequence.length = m_tx_data_length;
    m_current_sequence.repeats = 0;
    m_current_sequence.end_delay = 0;
    UNUSED_RETURN_VALUE(nrfx_pwm_simple_playback(&m_pwm, &m_current_sequence, 1, PWM_TRANSMITTER_SEQUENCE_FLAGS));    
    m_sequence_running = true;
}


/*
 * Profiling
 *****************************************************************************/


/**@brief Function for initializing profiling pins */
static void profile_pins_init(void)
{
#ifdef ADC_SAMPLER_PROFILING
    nrf_gpio_cfg_output(PROFILE_PIN_SAADC);
    nrf_gpio_cfg_output(PROFILE_PIN_SAADC_HANDLER);
    nrf_gpio_cfg_output(PROFILE_PIN_ALGORITHM);
    nrf_gpio_pin_set(PROFILE_PIN_SAADC);
    nrf_gpio_pin_clear(PROFILE_PIN_SAADC_HANDLER);
    nrf_gpio_pin_clear(PROFILE_PIN_ALGORITHM);

    nrf_gpiote_task_configure(PROFILE_SAADC_GPIOTE_TASK_IDX, PROFILE_PIN_SAADC, NRF_GPIOTE_POLARITY_TOGGLE, NRF_GPIOTE_INITIAL_VALUE_HIGH);
    nrf_gpiote_task_enable(PROFILE_SAADC_GPIOTE_TASK_IDX);
    nrf_ppi_channel_endpoint_setup(PROFILE_SAADC_EVENTS_END_PPI_CHANNEL,
        nrf_saadc_event_address_get(NRF_SAADC_EVENT_RESULTDONE),
        nrf_gpiote_task_addr_get(NRF_GPIOTE_TASKS_OUT_4));
    nrf_ppi_channel_enable(PROFILE_SAADC_EVENTS_END_PPI_CHANNEL);
#endif

#ifdef PWM_TRANSMITTER_PROFILING
    nrf_gpio_cfg_output(PROFILE_PIN_PWM);
    nrf_gpio_cfg_output(PROFILE_PIN_TX_FUNCTION);
    nrf_gpio_pin_clear(PROFILE_PIN_PWM);
    nrf_gpio_pin_clear(PROFILE_PIN_TX_FUNCTION);

    nrf_gpiote_task_configure(PROFILE_PWM_GPIOTE_TASK_IDX, PROFILE_PIN_PWM, NRF_GPIOTE_POLARITY_TOGGLE, NRF_GPIOTE_INITIAL_VALUE_LOW);
    nrf_gpiote_task_enable(PROFILE_PWM_GPIOTE_TASK_IDX);
    nrf_ppi_channel_endpoint_setup(PROFILE_PWM_EVENTS_SEQSTARTED_0_PPI_CHANNEL,
        nrf_pwm_event_address_get(m_pwm.p_registers, NRF_PWM_EVENT_SEQSTARTED0),
        nrf_gpiote_task_addr_get(NRFX_CONCAT_2(NRF_GPIOTE_TASKS_OUT_, PROFILE_PWM_GPIOTE_TASK_IDX)));
    nrf_ppi_channel_endpoint_setup(PROFILE_PWM_EVENTS_SEQSTARTED_1_PPI_CHANNEL,
        nrf_pwm_event_address_get(m_pwm.p_registers, NRF_PWM_EVENT_SEQSTARTED1),
        nrf_gpiote_task_addr_get(NRFX_CONCAT_2(NRF_GPIOTE_TASKS_OUT_, PROFILE_PWM_GPIOTE_TASK_IDX)));
    nrf_ppi_channel_endpoint_setup(PROFILE_PWM_EVENTS_SEQEND_0_PPI_CHANNEL,
        nrf_pwm_event_address_get(m_pwm.p_registers, NRF_PWM_EVENT_SEQEND0),
        nrf_gpiote_task_addr_get(NRFX_CONCAT_2(NRF_GPIOTE_TASKS_OUT_, PROFILE_PWM_GPIOTE_TASK_IDX)));
    nrf_ppi_channel_endpoint_setup(PROFILE_PWM_EVENTS_SEQEND_1_PPI_CHANNEL,
        nrf_pwm_event_address_get(m_pwm.p_registers, NRF_PWM_EVENT_SEQEND1),
        nrf_gpiote_task_addr_get(NRFX_CONCAT_2(NRF_GPIOTE_TASKS_OUT_, PROFILE_PWM_GPIOTE_TASK_IDX)));
    nrf_ppi_channel_enable(PROFILE_PWM_EVENTS_SEQSTARTED_0_PPI_CHANNEL);
    nrf_ppi_channel_enable(PROFILE_PWM_EVENTS_SEQSTARTED_1_PPI_CHANNEL);
    nrf_ppi_channel_enable(PROFILE_PWM_EVENTS_SEQEND_0_PPI_CHANNEL);
    nrf_ppi_channel_enable(PROFILE_PWM_EVENTS_SEQEND_1_PPI_CHANNEL);
#endif
}


/*****************************************************************************
 * Public API
 *****************************************************************************/


uint32_t nrf_vt_backend_init(nrf_vt_backend_init_t * p_config)
{
    nrfx_err_t err_code;

    if (p_config->rx_pin_p != NRF_VT_PIN_DISCONNECTED)
    {
        err_code = adc_sampler_init(p_config->baud_rate, p_config->threshold_milli_volts, p_config->saadc_interrupt_priority,
                                    p_config->rx_pin_n != NRF_VT_PIN_DISCONNECTED);
        if (err_code != NRF_SUCCESS)
        {
            return err_code;
        }
        err_code = adc_sampler_pin_init(p_config->rx_pin_p, p_config->rx_pin_n);
        if (err_code != NRF_SUCCESS)
        {
            adc_sampler_uninit();
            return err_code;
        }
    }

    if (p_config->tx_pin != NRF_VT_PIN_DISCONNECTED)
    {
        err_code = pwm_transmitter_init(p_config->baud_rate, p_config->tx_pin, p_config->pwm_interrupt_priority);
        if (err_code != NRF_SUCCESS)
        {
            if (p_config->rx_pin_p != NRF_VT_PIN_DISCONNECTED)
            {
                adc_sampler_pin_uninit();
                adc_sampler_uninit();
            }
            return err_code;
        }
    }

    m_rx_pin_p = p_config->rx_pin_p;
    m_rx_pin_n = p_config->rx_pin_n;
    m_tx_pin = p_config->tx_pin;

#if defined(ADC_SAMPLER_PROFILING) || defined(PWM_TRANSMITTER_PROFILING)
    profile_pins_init();
#endif

    return NRF_SUCCESS;
}


void nrf_vt_backend_uninit(void)
{
    if (m_rx_pin_p != NRF_VT_PIN_DISCONNECTED)
    {
        adc_sampler_pin_uninit();
        adc_sampler_uninit(true);
    }
    if (m_tx_pin != NRF_VT_PIN_DISCONNECTED)
    {
        pwm_transmitter_uninit(m_tx_pin);
    }

    m_rx_pin_p = NRF_VT_PIN_DISCONNECTED;
    m_rx_pin_n = NRF_VT_PIN_DISCONNECTED;
    m_tx_pin = NRF_VT_PIN_DISCONNECTED;
}


uint32_t nrf_vt_backend_rx_start(void)
{
    if (!nrfx_saadc_is_busy())
    {
        nrfx_saadc_calibrate_offset();
        return NRF_SUCCESS;
    }
    else
    {
        return NRF_ERROR_BUSY;
    }
}


void nrf_vt_backend_rx_stop(void)
{
    nrfx_saadc_abort();
}


bool nrf_vt_backend_rx_started(void)
{
    return nrfx_saadc_is_busy();
}


uint32_t nrf_vt_backend_tx_buffer_write_frame(const uint8_t const * p_data, uint32_t bits)
{
#ifdef PWM_TRANSMITTER_PROFILING
    nrf_gpio_pin_set(PROFILE_PIN_TX_FUNCTION);
#endif
    if (!m_sequence_running)
    {
        if (PWM_TRANSMITTER_TX_BUFFER_SIZE - m_tx_data_length > bits)
        {
            tx_buffer_write_frame(p_data, &m_tx_buffer[m_tx_data_length], bits);
            m_tx_data_length += _VT_START_BITS + bits + _VT_STOP_BITS + _VT_END_DELAY;
#ifdef PWM_TRANSMITTER_PROFILING
            nrf_gpio_pin_clear(PROFILE_PIN_TX_FUNCTION);
#endif
            return NRF_SUCCESS;
        }
        else
        {
#ifdef PWM_TRANSMITTER_PROFILING
            nrf_gpio_pin_clear(PROFILE_PIN_TX_FUNCTION);
#endif
            return NRF_ERROR_NO_MEM;
        }
    }
    else
    {
#ifdef PWM_TRANSMITTER_PROFILING
        nrf_gpio_pin_clear(PROFILE_PIN_TX_FUNCTION);
#endif
        return NRF_ERROR_BUSY;
    }
}


uint32_t nrf_vt_backend_tx_frame_cast_init(uint8_t cast_no, uint32_t payload_bits)
{
    if (cast_no >= PWM_TRANSMITTER_FRAME_CAST_COUNT || payload_bits > PWM_TRANSMITTER_MAX_PAYLOAD_BITS)
    {
        return NRF_ERROR_INVALID_PARAM;
    }
    frame_cast_init(cast_no, payload_bits);
    return NRF_SUCCESS;
}


uint32_t nrf_vt_backend_tx_buffer_write_cast_frame(uint8_t cast_no, const uint8_t const * p_data)
{
#ifdef PWM_TRANSMITTER_PROFILING
    nrf_gpio_pin_set(PROFILE_PIN_TX_FUNCTION);
#endif
    if (!m_sequence_running)
    {
        uint32_t bits = m_frame_payload_lengths[cast_no];

        if (bits > 0 && PWM_TRANSMITTER_TX_BUFFER_SIZE - m_tx_data_length > bits)
        {
            tx_buffer_write_cast_payload(cast_no, p_data, &m_tx_buffer[m_tx_data_length], bits);
            m_tx_data_length += _VT_START_BITS + bits + _VT_STOP_BITS + _VT_END_DELAY;
#ifdef PWM_TRANSMITTER_PROFILING
            nrf_gpio_pin_clear(PROFILE_PIN_TX_FUNCTION);
#endif
            return NRF_SUCCESS;
        }
        else
        {
#ifdef PWM_TRANSMITTER_PROFILING
            nrf_gpio_pin_clear(PROFILE_PIN_TX_FUNCTION);
#endif
            return NRF_ERROR_NO_MEM;
        }
    }
    else
    {
#ifdef PWM_TRANSMITTER_PROFILING
        nrf_gpio_pin_clear(PROFILE_PIN_TX_FUNCTION);
#endif
        return NRF_ERROR_BUSY;
    }
}


void nrf_vt_backend_tx_buffer_clear(void)
{
    if (!m_sequence_running)
    {
        m_tx_data_length = 0;
    }
}


uint32_t nrf_vt_backend_tx_start(void)
{
    if (!m_sequence_running)
    {
        if (m_tx_data_length > 0)
        {
             pwm_transmitter_start_sequence();
        }
        return NRF_SUCCESS;
    }
    else
    {
        return NRF_ERROR_BUSY;
    }
}


bool nrf_vt_backend_tx_started(void)
{
    return m_sequence_running;
}


void nrf_vt_backend_tx_stop(void)
{
    nrfx_pwm_stop(&m_pwm, true);

    m_sequence_running = false;
    m_tx_data_length = 0;
}

#endif // defined(NRF_VT_ENABLED) && NRF_VT_ENABLED
