#include "sdk_config.h"

#include "nrf_log.h"

#include <nrfx.h>

#include "nrf_vt_uart_frontend.h"
#include "nrf_vt_constants.h"
#include "nrf_vt_backend.h"

#if VT_FRONTEND_DEBUG || VT_FRONTEND_INDICATE_PACKET
#include "nrf_gpio.h"
#endif


#if defined(NRF_VT_ENABLED) && NRF_VT_ENABLED
#if defined(NRF_VT_FRONTEND) && NRF_VT_FRONTEND == 0


/*****************************************************************************
 * Constants
 *****************************************************************************/

/** Threshold value in millivolts for distinguishing 0 and 1 bits */
#ifndef NRF_VT_UART_PIN_N
#define THRESHOLD_VALUE_MILLIVOLTS  (1200)
#else
#define THRESHOLD_VALUE_MILLIVOLTS  (1600)
#endif

/** Amount of bits in a frame payload */
#define PAYLOAD_BITS                (8)
/** Frame cast number used */
#define FRAME_CAST_NUMBER           (0)


/*****************************************************************************
 * Profiling
 *****************************************************************************/

#if VT_FRONTEND_INDICATE_PACKET
#define VT_UART_PACKET_PIN  (22)
#endif

#if VT_FRONTEND_INDICATE_ERROR
#define VT_UART_ERROR_PIN   (23)
#endif


/*****************************************************************************
 * Static variables
 *****************************************************************************/

/** UART frame parser states */
typedef enum
{
    WAITING_FOR_START,
    IN_WORD,
    WAITING_FOR_STOP
}
nrf_vt_uart_parser_state_t;


/** The current state of the frame parser */
static nrf_vt_uart_parser_state_t   m_parser_state = WAITING_FOR_START;
/** The amount of bits read in the current frame */
static uint_fast8_t                 m_bit_count = 0;
/** Buffer for the payload of the current frame */
static uint_fast8_t                 m_buf_word = 0;

/** Fake errorsrc mask */
static uint32_t                     m_errorsrc = 0;

/** Context pointer to be passed to the application event handler */
static void *                       m_p_context = NULL;
/** Application event handler */
static nrf_vt_uart_event_handler_t  m_event_handler = NULL;

/** Pointer to the primary RX buffer (supplied by the user) */
static uint8_t *                    m_p_rx_buffer = NULL;
/** Maximum length of the primary RX buffer */
static size_t                       m_rx_buffer_length = 0;

/** Pointer to the secondary RX buffer (supplied by the user) */
static uint8_t *                    m_p_rx_secondary_buffer = NULL;
/** Maximum length of the secondary RX buffer */
static size_t                       m_rx_secondary_buffer_length = 0;

/** Amount of bytes stored so far in the current RX buffer */
static size_t                       m_rx_counter = 0;

/** Pointer to the TX buffer (supplied by the user) */
static uint8_t const *              m_p_tx_buffer = NULL;
/** Length of the TX buffer */
static size_t m_tx_buffer_length = 0;


/*****************************************************************************
 * Macros / static functions
 *****************************************************************************/

/**@brief Get the numerical baudrate value for a given UART_BAUDRATE constant */
static uint32_t actual_baudrate_get(nrf_uart_baudrate_t baud_rate)
{
    switch (baud_rate)
    {
        case NRF_UART_BAUDRATE_1200:
            return 1200;
        case NRF_UART_BAUDRATE_2400:
            return 2400;
        case NRF_UART_BAUDRATE_4800:
            return 4800;
        case NRF_UART_BAUDRATE_9600:
            return 9600;
        case NRF_UART_BAUDRATE_14400:
            return 14400;
        case NRF_UART_BAUDRATE_19200:
            return 19200;
        case NRF_UART_BAUDRATE_28800:
            return 28800;
        case NRF_UART_BAUDRATE_31250:
            return 31250;
        case NRF_UART_BAUDRATE_38400:
            return 38400;
        case NRF_UART_BAUDRATE_56000:
            return 56000;
        case NRF_UART_BAUDRATE_57600:
            return 57600;
        case NRF_UART_BAUDRATE_76800:
            return 76800;
        case NRF_UART_BAUDRATE_115200:
            return 115200;
        case NRF_UART_BAUDRATE_230400:
            return 230400;
        case NRF_UART_BAUDRATE_250000:
            return 250000;
        case NRF_UART_BAUDRATE_460800:
            return 460800;
        case NRF_UART_BAUDRATE_921600:
            return 921600;
        case NRF_UART_BAUDRATE_1000000:
            return 1000000;
        default:
            return 0;
    }
}


/**@brief Send an error event to the application */
static void rx_error(uint32_t errorsrc)
{
#if VT_FRONTEND_INDICATE_ERROR
    nrf_gpio_pin_toggle(VT_UART_ERROR_PIN);
#endif

    if (m_event_handler != NULL)
    {
        nrf_vt_uart_event_t event;

        event.type = NRF_VT_UART_EVT_ERROR;

        event.data.error.error_mask = errorsrc;
        event.data.error.rxtx.p_data = m_p_rx_buffer;
        event.data.error.rxtx.bytes = m_rx_counter;

        m_p_rx_buffer = m_p_rx_secondary_buffer;
        m_rx_buffer_length = m_rx_secondary_buffer_length;
        m_rx_secondary_buffer_length = 0;
        m_rx_counter = 0;

        m_errorsrc = errorsrc;
        m_event_handler(&event, m_p_context);
        m_errorsrc = 0;
    }
    else
    {
        m_p_rx_buffer = m_p_rx_secondary_buffer;
        m_rx_buffer_length = m_rx_secondary_buffer_length;
        m_rx_secondary_buffer_length = 0;
        m_rx_counter = 0;
    }
}


/**@brief Send a BREAK error event to the application */
static inline void __attribute__((always_inline)) rx_break_error(void)
{
#if VT_FRONTEND_INDICATE_ERROR
    // NRF_LOG_ERROR("Framing error on word %0X", m_buf_word);
#endif
    rx_error(NRF_UART_ERROR_OVERRUN_MASK | NRF_UART_ERROR_FRAMING_MASK);
}


/**@brief Send a FRAMING error event to the application */
static inline void __attribute__((always_inline)) rx_framing_error(void)
{
#if VT_FRONTEND_INDICATE_ERROR
    // NRF_LOG_ERROR("Framing error on word %0X", m_buf_word);
#endif
    rx_error(NRF_UART_ERROR_FRAMING_MASK);
}


// TODO: This should be generated when there is a parity error
/**@brief Send a PARITY error event to the application */
static inline void __attribute__((always_inline)) rx_parity_error(void)
{
    rx_error(NRF_UART_ERROR_PARITY_MASK);
}


/**@brief Send an OVERRUN error event to the application */
static inline void __attribute__((always_inline))  rx_overrun_error(void)
{
    rx_error(NRF_UART_ERROR_OVERRUN_MASK);
}


/**@brief Send an RX_DONE event to the application */
static void rx_done(void)
{
    if (m_event_handler != NULL)
    {
        nrf_vt_uart_event_t event;

        event.type = NRF_VT_UART_EVT_RX_DONE;
        event.data.rxtx.p_data = m_p_rx_buffer;
        event.data.rxtx.bytes = m_rx_counter;

        m_p_rx_buffer = m_p_rx_secondary_buffer;
        m_rx_buffer_length = m_rx_secondary_buffer_length;
        m_rx_secondary_buffer_length = 0;
        m_rx_counter = 0;

        m_event_handler(&event, m_p_context);
    }
    else
    {
        m_p_rx_buffer = m_p_rx_secondary_buffer;
        m_rx_buffer_length = m_rx_secondary_buffer_length;
        m_rx_secondary_buffer_length = 0;
        m_rx_counter = 0;
    }
}


/**@brief Copy the next byte into the RX buffer */
static void rx_byte(uint8_t word)
{
#if VT_FRONTEND_INDICATE_PACKET
    nrf_gpio_pin_toggle(VT_UART_PACKET_PIN);
#endif
    if (m_rx_buffer_length == 0)
    {
        // Do nothing
    }
    else
    {
        m_p_rx_buffer[m_rx_counter] = word;
        m_rx_counter++;
        if (m_rx_counter == m_rx_buffer_length)
        {
            rx_done();
        }
    }
}


/**@brief Send a TX_DONE event to the application */
static void tx_done(void)
{
    if (m_event_handler != NULL)
    {
        nrf_vt_uart_event_t event;

        event.type = NRF_VT_UART_EVT_TX_DONE;
        event.data.rxtx.p_data = (uint8_t *) m_p_tx_buffer;
        event.data.rxtx.bytes = m_tx_buffer_length;

        m_event_handler(&event, m_p_context);
    }

    m_p_tx_buffer = NULL;
    m_tx_buffer_length = 0;
}


/*****************************************************************************
 * Backend API
 *****************************************************************************/

#if VT_FRONTEND_INDICATE_ERROR || VT_FRONTEND_INDICATE_PACKET
static void profiling_pins_init(void)
{
#if VT_FRONTEND_ERROR
    nrf_gpio_cfg_output(VT_UART_ERROR_PIN);
    nrf_gpio_pin_clear(VT_UART_ERROR_PIN);
#endif
#if VT_FRONTEND_INDICATE_PACKET
    nrf_gpio_cfg_output(VT_UART_PACKET_PIN);
    nrf_gpio_pin_set(VT_UART_PACKET_PIN);
#endif
}
#endif


/*****************************************************************************
 * Backend API
 *****************************************************************************/

/**@brief Reset the state of the parser */
void vt_frontend_parser_state_reset(void)
{
    m_bit_count = 0;
    m_buf_word = 0;
    m_parser_state = WAITING_FOR_START;
}


/**@brief Parse the next bit; return false if the backend should wait for the next start bit */
#ifdef VT_FRONTEND_DEBUG
uint32_t vt_frontend_parse_next_bit(uint_fast8_t bit)
#else
bool vt_frontend_parse_next_bit(uint_fast8_t bit)
#endif
{
    switch (m_parser_state)
    {
        case WAITING_FOR_START:
        {
            if (bit == _VT_START_BIT_VALUE)
            {
                m_parser_state = IN_WORD;
#ifdef VT_FRONTEND_DEBUG
                return 0;
#else
                return true;
#endif
            }
            else
            {
                break;
            }
        }

        case IN_WORD:
        {
            // TODO: This should ideally not be run 8 times...
            m_buf_word |= (bit << m_bit_count);
            m_bit_count++;

            if (m_bit_count == PAYLOAD_BITS)
            {
                m_parser_state = WAITING_FOR_STOP;
            }
#ifdef VT_FRONTEND_DEBUG
            return 0;
#else
            return true;
#endif
        }

        case WAITING_FOR_STOP:
        {
            if (bit == _VT_STOP_BIT_VALUE)
            {
                rx_byte(m_buf_word);
            }
            else
            {
                // FIXME: Not consistent with the nRF UART as that peripheral generates 
                //        a FRAMING first and then a BREAK if the next bit
                //        is also 0
                if (m_buf_word == 0)
                {
                    rx_break_error();
#ifdef VT_FRONTEND_DEBUG
                    vt_frontend_parser_state_reset();
                    return 3;
#endif
                }
                else
                {
                    rx_framing_error();
#ifdef VT_FRONTEND_DEBUG
                    vt_frontend_parser_state_reset();
                    return 2;
#endif
                }
            }

            vt_frontend_parser_state_reset();

            break;
        }

        default:
        {
            break;
        }
    }
#ifdef VT_FRONTEND_DEBUG
    return 1;
#else
    return false;
#endif
}


/**@brief Function for handling buffer overruns */
void vt_frontend_overrun_handler(void)
{
    vt_frontend_parser_state_reset();
    rx_overrun_error();
}


/**@Brief Function for handling a finished transmission */
void vt_frontend_tx_done_handler(void)
{
    tx_done();
}


/*****************************************************************************
 * Public API
 *****************************************************************************/


nrfx_err_t nrf_vt_uart_init(nrf_vt_uart_t const *        p_instance,
                            nrf_vt_uart_config_t const * p_config,
                            nrf_vt_uart_event_handler_t  event_handler)
{
    UNUSED_PARAMETER(p_instance);

    uint32_t err_code;
    
    nrf_vt_backend_init_t backend_config = {
        .baud_rate = actual_baudrate_get(p_config->baudrate),
        .rx_pin_p = p_config->pselrxd != NRF_UART_PSEL_DISCONNECTED ? p_config->pselrxd : NRF_VT_PIN_DISCONNECTED,
        .tx_pin = p_config->pseltxd != NRF_UART_PSEL_DISCONNECTED ? p_config->pseltxd : NRF_VT_PIN_DISCONNECTED,
        .pwm_interrupt_priority = p_config->interrupt_priority,
        .saadc_interrupt_priority = p_config->interrupt_priority,
        .threshold_milli_volts = 1600 // THRESHOLD_VALUE_MILLIVOLTS
    };

#ifdef NRF_VT_UART_PIN_N
    backend_config.rx_pin_n = NRF_VT_UART_PIN_N;
#endif

    // TODO: For now the other UART struct members are ignored
    // TODO: Implement parity and HWFC
    
    err_code = nrf_vt_backend_init(&backend_config);

    switch (err_code)
    {
        case NRF_ERROR_INVALID_PARAM:
        {
            // Not ideal as the NRFX_ERROR_INVALID_STATE indicates a multiple init error, but we want the
            // return codes the same as the nrfx_uart(e)
            return NRFX_ERROR_INVALID_STATE;
        }
        case NRF_ERROR_INVALID_STATE:
        {
            return NRFX_ERROR_INVALID_STATE;
        }
        case NRF_ERROR_BUSY:
        {
            return NRFX_ERROR_BUSY;
        }
        
        default: break;
    }

    err_code = nrf_vt_backend_tx_frame_cast_init(FRAME_CAST_NUMBER, PAYLOAD_BITS);
    if (err_code != NRF_SUCCESS)  // A bit complicated to avoid the above call from being optimized out
    {
        ASSERT(false);
    }

    m_event_handler = event_handler;
    m_p_context = p_config->p_context;

#if VT_FRONTEND_DEBUG || VT_FRONTEND_INDICATE_PACKET
    profiling_pins_init();
#endif

    return NRFX_SUCCESS;
}


#if NRF_VT_REPLACE_NRFX_DRIVER
nrfx_err_t nrfx_uarte_init(nrfx_uarte_t const *        p_instance,
                           nrfx_uarte_config_t const * p_config,
                           nrfx_uarte_event_handler_t  event_handler)
{
    return nrf_vt_uart_init((nrf_vt_uart_t const *) p_instance,
                            (nrf_vt_uart_config_t const *) p_config,
                            (nrf_vt_uart_event_handler_t) event_handler);
}
#endif


void nrf_vt_uart_uninit(nrf_vt_uart_t const * p_instance)
{
    UNUSED_PARAMETER(p_instance);

    nrf_vt_backend_uninit();

    vt_frontend_parser_state_reset();

    m_event_handler = NULL;
    m_p_context = NULL;

    m_p_rx_buffer = NULL;
    m_rx_buffer_length = 0;

    m_p_rx_secondary_buffer = NULL;
    m_rx_secondary_buffer_length = 0;

    m_rx_counter = 0;

    m_p_tx_buffer = NULL;
    m_tx_buffer_length = 0;
}


#if NRF_VT_REPLACE_NRFX_DRIVER
void nrfx_uarte_uninit(nrfx_uarte_t const * p_instance)
{
    nrf_vt_uart_uninit((nrf_vt_uart_t const *) p_instance);
}
#endif


nrfx_err_t nrf_vt_uart_tx(nrf_vt_uart_t const * p_instance,
                          uint8_t const *       p_data,
                          size_t                length)
{
    UNUSED_PARAMETER(p_instance);

    uint32_t err_code;

    if (!nrf_vt_uart_tx_in_progress(p_instance))
    {
        m_p_tx_buffer = p_data;
        m_tx_buffer_length = length;

        for (size_t i = 0; i < length; i++)
        {
            err_code = nrf_vt_backend_tx_buffer_write_cast_frame(FRAME_CAST_NUMBER, &p_data[i]);
            if (err_code != NRF_SUCCESS)
            {
                nrf_vt_backend_tx_buffer_clear();
                
                // Some relevant errors such as buffer exhaustion are unfortunately hidden here,
                // but we only have two errors to work with given the nrfx_uart(e) interface
                return NRF_ERROR_BUSY;
            }
        }
        if (nrf_vt_backend_tx_start() == NRF_ERROR_BUSY)
        {
            return NRFX_ERROR_BUSY;
        }
        return NRFX_SUCCESS;
    }
    else
    {
        return NRF_ERROR_BUSY;
    }
}


#if NRF_VT_REPLACE_NRFX_DRIVER
nrfx_err_t nrfx_uarte_tx(nrfx_uarte_t const * p_instance,
                         uint8_t const *      p_data,
                         size_t               length)
{
    return nrf_vt_uart_tx((nrf_vt_uart_t const *) p_instance,
                          p_data, length);
}
#endif


bool nrf_vt_uart_tx_in_progress(nrf_vt_uart_t const * p_instance)
{
    UNUSED_PARAMETER(p_instance);

    return nrf_vt_backend_tx_started();
}


#if NRF_VT_REPLACE_NRFX_DRIVER
bool nrfx_uarte_tx_in_progress(nrfx_uarte_t const * p_instance)
{
    return nrf_vt_uart_tx_in_progress((nrf_vt_uart_t const *) p_instance);
}
#endif


void nrf_vt_uart_tx_abort(nrf_vt_uart_t const * p_instance)
{
    UNUSED_PARAMETER(p_instance);

    nrf_vt_backend_tx_stop();

    m_p_tx_buffer = NULL;
    m_tx_buffer_length = 0;
}


#if NRF_VT_REPLACE_NRFX_DRIVER
void nrfx_uarte_tx_abort(nrfx_uarte_t const * p_instance)
{
    nrf_vt_uart_tx_abort((nrf_vt_uart_t const *) p_instance);
}
#endif


nrfx_err_t nrf_vt_uart_rx(nrf_vt_uart_t const * p_instance,
                          uint8_t *             p_data,
                          size_t                length)
{
    UNUSED_PARAMETER(p_instance);

    if (m_rx_buffer_length == 0)
    {
        m_p_rx_buffer = p_data;
        m_rx_buffer_length = length;
        m_rx_secondary_buffer_length = 0;
        m_rx_counter = 0;
    }
    else if (m_rx_secondary_buffer_length == 0)
    {
        m_p_rx_secondary_buffer = p_data;
        m_rx_secondary_buffer_length = length;
    }
    else
    {
        return NRFX_ERROR_BUSY;
    }

    if (!nrf_vt_backend_rx_started())
    {
        if (nrf_vt_backend_rx_start() == NRF_ERROR_BUSY)
        {
            return NRFX_ERROR_BUSY;
        }
    }

    return NRFX_SUCCESS;
}


#if NRF_VT_REPLACE_NRFX_DRIVER
nrfx_err_t nrfx_uarte_rx(nrfx_uarte_t const * p_instance,
                         uint8_t *            p_data,
                         size_t               length)
{
    return nrf_vt_uart_rx((nrf_vt_uart_t const *) p_instance,
                          p_data, length);
}
#endif


bool nrf_vt_uart_rx_ready(nrf_vt_uart_t const * p_instance)
{
    UNUSED_PARAMETER(p_instance);

    // TODO: Implement if/when blocking mode is added

    return false;
}


#if NRF_VT_REPLACE_NRFX_DRIVER
bool nrfx_uarte_rx_ready(nrfx_uarte_t const * p_instance)
{
    return nrf_vt_uart_rx_ready((nrf_vt_uart_t const *) p_instance);
}
#endif


void nrf_vt_uart_rx_enable(nrf_vt_uart_t const * p_instance)
{
    UNUSED_PARAMETER(p_instance);

    nrf_vt_backend_rx_start();
}


void nrf_vt_uart_rx_disable(nrf_vt_uart_t const * p_instance)
{
    UNUSED_PARAMETER(p_instance);

    nrf_vt_backend_rx_stop();
}


void nrf_vt_uart_rx_abort(nrf_vt_uart_t const * p_instance)
{
    UNUSED_PARAMETER(p_instance);

    nrf_vt_backend_rx_stop();
}


#if NRF_VT_REPLACE_NRFX_DRIVER
void nrfx_uarte_rx_abort(nrfx_uarte_t const * p_instance)
{
    nrf_vt_uart_rx_abort((nrf_vt_uart_t const *) p_instance);
}
#endif


uint32_t nrf_vt_uart_errorsrc_get(nrf_vt_uart_t const * p_instance)
{
    UNUSED_PARAMETER(p_instance);

    uint32_t errorsrc = m_errorsrc;
    m_errorsrc = 0;
    return errorsrc;
}


#if NRF_VT_REPLACE_NRFX_DRIVER
uint32_t nrfx_uarte_errorsrc_get(nrfx_uarte_t const * p_instance)
{
    return nrf_vt_uart_errorsrc_get((nrf_vt_uart_t const *) p_instance);
}
#endif


#endif // defined(NRF_VT_FRONTEND) && NRF_VT_FRONTEND == 0
#endif // defined(NRF_VT_ENABLED) && NRF_VT_ENABLED