#include "nrf_log.h"
#include "nrf_ringbuf.h"
#include "app_error.h"

#include "nrf_vt_uart_frontend.h"

#include "aci.h"

#define VT_ACI_ECHO_DEBUG 1

#define ACI_RINGBUF_PACKETS 8
#define ACI_RINGBUF_SIZE (ACI_RINGBUF_PACKETS * sizeof(vt_aci_packet_t))

NRF_RINGBUF_DEF(m_aci_packets, ACI_RINGBUF_SIZE);


#define VT_ACI_PACKET_LENGTH_VALID(length) (1 <= length && length < VT_ACI_PDU_MAX_LENGTH + 2)


typedef enum
{
    WAITING_FOR_LENGTH,
    WAITING_FOR_OPCODE,
    WAITING_FOR_PAYLOAD
}
vt_aci_parser_state_t;


static vt_aci_parser_state_t m_parser_state = WAITING_FOR_LENGTH;
static uint_fast8_t m_pdu_counter = 0;

static vt_aci_packet_t m_ingoing_packet;

static vt_aci_evt_handler_t m_event_handler = NULL;

#if VT_ACI_ECHO_DEBUG
static uint8_t m_ingoing_echo_counter = 0;
static uint8_t m_outgoing_echo_counter = 0;
#endif


static void aci_internal_handler(vt_aci_packet_t * p_packet)
{
    if (p_packet->opcode == (uint8_t) VT_ACI_CMD_ECHO)
    {
#if VT_ACI_ECHO_DEBUG
        NRF_LOG_INFO("Got ECHO command. (%d) (len: %d)", m_ingoing_echo_counter, p_packet->length);
        m_ingoing_echo_counter++;
#endif
        vt_aci_send(VT_ACI_EVT_ECHO, p_packet->length - 1, p_packet->pdu);
    }

    else
    {
        NRF_LOG_INFO("Got command 0x%X", p_packet->opcode);
    }
}


static void rx_next(uint8_t value)
{
    switch (m_parser_state)
    {
        case WAITING_FOR_LENGTH:
        {
            if (VT_ACI_PACKET_LENGTH_VALID(value))
            {
                m_ingoing_packet.length = value;
                m_parser_state = WAITING_FOR_OPCODE;
            }
            break;
        }

        case WAITING_FOR_OPCODE:
        {
            m_ingoing_packet.opcode = value;
            
            if (m_ingoing_packet.length == 1)
            {
                aci_internal_handler(&m_ingoing_packet);
                if (m_event_handler != NULL)
                {
                    m_event_handler(&m_ingoing_packet);
                }
                m_parser_state = WAITING_FOR_LENGTH;
            }
            else
            {
                m_parser_state = WAITING_FOR_PAYLOAD;
            }
            
            break;
        }

        case WAITING_FOR_PAYLOAD:
        {
            m_ingoing_packet.pdu[m_pdu_counter] = value;
            m_pdu_counter++;

            if (m_pdu_counter + 1 == m_ingoing_packet.length)
            {
                aci_internal_handler(&m_ingoing_packet);
                if (m_event_handler != NULL)
                {
                    m_event_handler(&m_ingoing_packet);
                }

                m_pdu_counter = 0;
                m_parser_state = WAITING_FOR_LENGTH;
            }

            break;
        }

        default:
        {
            break;
        }
    }
}


static uint8_t m_rx_buf_a[1];
static uint8_t m_rx_buf_b[1];


static void tx_attempt(void)
{
    uint32_t err_code;

    if (!nrf_vt_uart_tx_in_progress(NULL))
    {
        vt_aci_packet_t * p_packet;
        size_t length = sizeof(vt_aci_packet_t);

        if (nrf_ringbuf_get(&m_aci_packets, (uint8_t **)&p_packet, &length, true) == NRF_SUCCESS)
        {
            if (length >= sizeof(vt_aci_packet_t))
            {
                err_code = nrf_vt_uart_tx(NULL, (uint8_t *)p_packet, p_packet->length + 1);
                if (err_code == NRF_SUCCESS)
                {
#if VT_ACI_ECHO_DEBUG
                    NRF_LOG_INFO("Sent EchoEvent.   (%d)", m_outgoing_echo_counter);
                    m_outgoing_echo_counter++;
#endif
                    nrf_ringbuf_free(&m_aci_packets, length);
                }
            }
        }
    }
}


static void uart_event_handler(nrf_vt_uart_event_t const * p_event, void * p_context)
{
    UNUSED_PARAMETER(p_context);

    switch (p_event->type)
    {
        case NRF_VT_UART_EVT_TX_DONE: // Requested TX transfer completed.
        {
            tx_attempt();
            break;
        }
        
        case NRF_VT_UART_EVT_RX_DONE: // Requested RX transfer completed.
        {
            rx_next(*(p_event->data.rxtx.p_data));
            nrf_vt_uart_rx(NULL, p_event->data.rxtx.p_data, 1);
            break;
        }
        
        case NRF_VT_UART_EVT_ERROR:  // Error reported by UART peripheral.
        {
            NRF_LOG_INFO("UART Error: 0x%X", p_event->data.error.error_mask);
            nrf_vt_uart_rx(NULL, p_event->data.error.rxtx.p_data, 1);
            break;
        }

        default:
        {
            break;
        }
    }
}



uint32_t vt_aci_init(void * p_uart_config, vt_aci_evt_handler_t evt_handler)
{
    APP_ERROR_CHECK(nrf_vt_uart_init(NULL, (nrf_vt_uart_config_t *) p_uart_config, uart_event_handler));

    APP_ERROR_CHECK(nrf_vt_uart_rx(NULL, &m_rx_buf_a[0], 1));
    APP_ERROR_CHECK(nrf_vt_uart_rx(NULL, &m_rx_buf_b[0], 1));

    nrf_ringbuf_init(&m_aci_packets);

    m_event_handler = evt_handler;

    return NRF_SUCCESS;
}


uint32_t vt_aci_send(vt_aci_cmd_t opcode, uint8_t data_length, uint8_t * p_data)
{
    uint32_t err_code;
    
    vt_aci_packet_t packet = {
        .length = sizeof(uint8_t) * (/* Opcode length */ 1 + data_length),
        .opcode = (uint8_t)opcode,
        .pdu = {0}
    };

    memcpy(&packet.pdu, p_data, data_length);

    size_t length = sizeof(packet);

    nrf_ringbuf_cpy_put(&m_aci_packets, (const uint8_t *) &packet, &length);

    tx_attempt();
}
